import os
import sys

# Add the parent directory to the path so we can import the module
sys.path.append(
    os.path.realpath(
        os.path.join(os.path.realpath(__file__), "..", "..", "json_lineage")
    )
)



from unittest import TestCase

from json_lineage import exceptions


class BinaryExecutionException(TestCase):
    def test__str__returns_a_string(self):
        """Test that the `__str__` method returns a string."""
        err = exceptions.BinaryExecutionException(b"error")
        self.assertIsInstance(str(err), str)



import asyncio
import os
import platform
import subprocess
import typing as _t
from collections.abc import Awaitable, Coroutine

from .exceptions import BinaryExecutionException

__all__ = [
    "BinaryReader",
    "AsyncBinaryReader",
]


def get_bin_path() -> str:
    """Get the path to the jsonl_converter binary."""
    bin_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "bin")
    if platform.system() == "Windows":
        return os.path.join(bin_dir, "jsonl_converter.exe")
    else:
        return os.path.join(bin_dir, "jsonl_converter")


class BaseBinaryReader:
    """Base class for the `BinaryReader` and `AsyncBinaryReader` classes."""

    def __init__(self, filepath: str):
        self.bin_path = get_bin_path()
        self.file_path = filepath
        self._proc: _t.Optional[
            _t.Union[subprocess.Popen, asyncio.subprocess.Process]
        ] = None

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} bin_path={self.bin_path} "
            f"file_path={self.file_path}>"
        )

    def kill_subprocess_proc(self) -> None:
        """Kill the subprocess process."""
        if self._proc is None:
            return

        for stream in ("stdout", "stderr"):
            stream = getattr(self._proc, stream)
            if hasattr(stream, "close"):
                stream.close()

        try:
            self._proc.terminate()
        except ProcessLookupError:
            pass
        self._proc = None


class BinaryReader(BaseBinaryReader):
    """Subprocess wrapper for the jsonl_converter binary."""

    def __iter__(self):
        return BinaryIterator(self.popen())

    def popen(self) -> subprocess.Popen:
        """Run the binary and return a Popen object."""
        self._proc = subprocess.Popen(
            [self.bin_path, self.file_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        return self._proc


class BinaryIterator:
    """Iterator for the `BinaryReader` class."""

    def __init__(self, process: subprocess.Popen):
        self.process = process

    def __iter__(self):
        return self

    def __next__(self) -> str:
        self.raise_err_if_stderr()

        if self.process.stdout is None:
            raise StopIteration
        line = self.process.stdout.readline().strip()
        self.raise_err_if_stderr()
        if not line and self.process.poll() is not None:
            raise StopIteration

        return line

    def raise_err_if_stderr(self) -> None:
        """Raise an exception if the process has exited with a non-zero
        code.
        """
        if self.process.poll() is not None and self.process.poll() != 0:
            if self.process.stderr is None:
                raise BinaryExecutionException(
                    f"Process exited with code {self.process.poll()}"
                )
            else:
                raise BinaryExecutionException(self.process.stderr.read())


class AsyncBinaryReader(BaseBinaryReader):
    """Async subprocess wrapper for the jsonl_converter binary."""

    async def popen(self) -> asyncio.subprocess.Process:
        """Run the binary and return a Popen object."""
        self._proc = await asyncio.create_subprocess_exec(
            self.bin_path,
            self.file_path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return self._proc

    async def read_output(self, process: asyncio.subprocess.Process) -> str:
        if process.stdout is None:
            return ""

        line = await process.stdout.readline()
        if not line:
            return ""

        return line.decode().rstrip()

    def __aiter__(self):
        return AsyncBinaryIterator(self.popen())


class AsyncBinaryIterator:
    def __init__(self, process_coro: Coroutine):
        self.process_coro = process_coro
        self.process: asyncio.subprocess.Process | None = None

    async def __aiter__(self):
        return self

    async def __anext__(self) -> str:
        if self.process is None:
            self.process = await _t.cast(
                Awaitable[asyncio.subprocess.Process],
                self.process_coro,
            )
        output = await self.read_output(self.process)
        await self.raise_err_if_stderr()

        if not output:
            raise StopAsyncIteration

        return output

    async def read_output(self, process: asyncio.subprocess.Process) -> str:
        if process.stdout is None:
            return ""
        line = await process.stdout.readline()
        if not line:
            return ""
        return line.decode().rstrip()

    async def raise_err_if_stderr(self):
        if (
            self.process.returncode is not None
            and self.process.returncode != 0
        ):
            err = await self.process.stderr.read()
            if err:
                raise BinaryExecutionException(err)
            else:
                raise BinaryExecutionException(
                    f"Process exited with code {self.process.returncode}"
                )



"""This module contains the command line interface for the json_lineage
adapter.
"""

import argparse

from .bin_interface import BinaryReader


def parse_args() -> argparse.Namespace:
    """Parses the command line arguments."""
    parser = argparse.ArgumentParser(
        prog="python -m json_lineage",
        description="Read and convert JSON to JSONL (JSON Lines) format.",
    )
    parser.add_argument(
        "filepath",
        type=str,
        help="Path to the JSON file to read.",
    )
    parser.add_argument(
        "--output-file",
        "-o",
        type=str,
        help="Path to the output file.",
    )

    return parser.parse_args()


def print_lines(reader: BinaryReader) -> None:
    """Prints the lines from the given reader to stdout."""
    for line in reader:
        print(line)


def write_lines(reader: BinaryReader, filepath: str) -> None:
    """Writes the lines from the given reader to the given filepath."""
    with open(filepath, "w") as f:
        for line in reader:
            f.write(line + "\n")


def main() -> None:
    """The main entrypoint for the json_lineage adapter when run as a
    module.
    """
    args = parse_args()
    reader = BinaryReader(args.filepath)
    if args.output_file:
        write_lines(reader, args.output_file)
    else:
        print_lines(reader)



from unittest import TestCase

from json_lineage import aload, load
from json_lineage.bin_interface import AsyncBinaryReader, BinaryReader


class TestPublic(TestCase):
    """Tests for the `public` module."""

    def test_load_returns_binary_reader(self):
        """Test that `load` returns a `BinaryReader` object."""
        self.assertIsInstance(load("foo"), BinaryReader)

    def test_aload_returns_async_binary_reader(self):
        """Test that `aload` returns an `AsyncBinaryReader` object."""
        self.assertIsInstance(aload("foo"), AsyncBinaryReader)



import builtins
import sys
import tempfile
from unittest import TestCase
from unittest.mock import patch

from json_lineage import cli
from json_lineage.bin_interface import BinaryReader

from .helpers import SAMPLE_DATA_PATH


class TestParseArgs(TestCase):
    """Tests for the `parse_args` function."""

    def test_args_are_parsed(self):
        args = ["file.json", "-o", "file.jsonl"]
        with patch.object(sys, "argv", ["json_lineage"] + args):
            parsed_args = cli.parse_args()
            self.assertEqual(parsed_args.filepath, args[0])
            self.assertEqual(parsed_args.output_file, args[2])


class TestPrintLines(TestCase):
    """Tests for the `print_lines` function."""

    def test_print_lines(self):
        """Test that the `print_lines` function prints the lines from the
        given reader to stdout.
        """
        reader = BinaryReader(SAMPLE_DATA_PATH)
        with patch.object(builtins, "print") as mock_print:
            cli.print_lines(reader)
            self.assertEqual(mock_print.call_count, 2)
        reader.kill_subprocess_proc()

    def test_write_lines(self):
        """Test that the `write_lines` function writes the lines from the
        given reader to the given filepath.
        """
        reader = BinaryReader(SAMPLE_DATA_PATH)
        with tempfile.NamedTemporaryFile() as f:
            cli.write_lines(reader, f.name)
            self.assertEqual(f.read().count(b"\n"), 2)
        reader.kill_subprocess_proc()


class TestMain(TestCase):
    """Tests for the `main` function."""

    @patch.object(sys, "argv", ["json_lineage", SAMPLE_DATA_PATH])
    def test_main_with_no_output_file_prints(self):
        """Test that the `main` function prints the lines from the given
        reader to stdout when no output file is given.
        """
        with patch.object(cli, "print_lines") as mock_print_lines:
            cli.main()
            self.assertEqual(mock_print_lines.call_count, 1)

    @patch.object(
        sys,
        "argv",
        ["json_lineage", SAMPLE_DATA_PATH, "-o", "file.jsonl"],
    )
    def test_main_with_output_file_writes(self):
        """Test that the `main` function writes the lines from the given
        reader to the given filepath when an output file is given.
        """
        with patch.object(cli, "write_lines") as mock_write_lines:
            cli.main()
            self.assertEqual(mock_write_lines.call_count, 1)



import asyncio
import platform
import subprocess
import time
from types import SimpleNamespace
from typing import Callable, Union
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import patch

from json_lineage import bin_interface
from json_lineage.exceptions import BinaryExecutionException

from .helpers import SAMPLE_DATA_PATH


class ReaderInstanceMixin:
    create_reader_instance: Callable[
        ..., Union[bin_interface.BinaryReader, bin_interface.AsyncBinaryReader]
    ]

    def setUp(self):
        self.reader = self.create_reader_instance()

    def tearDown(self):
        self.reader.kill_subprocess_proc()


class TestGetBinPath(TestCase):
    """Tests for the `get_bin_path` function."""

    @patch.object(platform, "system", return_value="Windows")
    def test_get_bin_path_as_windows(self, _):
        """Test that the `get_bin_path` function returns the correct path on
        Windows.
        """
        self.assertTrue(
            bin_interface.get_bin_path().endswith("jsonl_converter.exe"),
        )

    @patch.object(platform, "system", return_value="Linux")
    def test_get_bin_path_as_linux(self, _):
        """Test that the `get_bin_path` function returns the correct path on
        Linux.
        """
        self.assertTrue(
            bin_interface.get_bin_path().endswith("jsonl_converter"),
        )


class TestBaseBinaryReader(TestCase):
    def test__init__works(self):
        """Test that the `BinaryReader` class can be instantiated."""
        bin_interface.BinaryReader("filepath")

    def test_repr(self):
        """Test that the `__repr__` method returns a string."""
        reader = bin_interface.BinaryReader("filepath")
        self.assertIsInstance(repr(reader), str)

    def test_kill_subprocess_proc_returns_if_no_proc(self):
        """Test that the `kill_subprocess_proc` method returns if there is no
        subprocess.
        """
        reader = bin_interface.BinaryReader("filepath")
        reader.kill_subprocess_proc()

    def test_kill_subprocess_proc_closes_files_and_terminates_proc(self):
        """Test that the `kill_subprocess_proc` method closes the files and
        terminates the subprocess.
        """
        reader = bin_interface.BinaryReader(SAMPLE_DATA_PATH)
        proc = reader.popen()
        reader.kill_subprocess_proc()
        self.assertTrue(proc.stdout.closed)
        self.assertTrue(proc.stderr.closed)
        time.sleep(0.01)
        self.assertIsNotNone(proc.poll())


class TestBinaryReader(ReaderInstanceMixin, TestCase):
    """Tests for the `BinaryReader` class."""

    @staticmethod
    def create_reader_instance():
        return bin_interface.BinaryReader(SAMPLE_DATA_PATH)

    def test_popen_returns_popen(self):
        """Test that the `popen` method returns a `subprocess.Popen` object."""
        proc = self.reader.popen()
        self.assertIsInstance(proc, subprocess.Popen)
        proc.communicate()

    def test_ppopen_readable_stdout(self):
        """Test that the `popen` method returns a `subprocess.Popen` object
        with a readable stdout.
        """
        proc = self.reader.popen()
        self.assertIsNotNone(proc.stdout.readline())
        proc.communicate()

    def test_iter(self):
        """Test that the `__iter__` method returns a `BinaryIterator`
        object.
        """
        self.assertIsInstance(iter(self.reader), bin_interface.BinaryIterator)
        self.reader._proc.communicate()

    def test_iter_next_valid(self):
        """Test that the `__next__` method iterates over the binary stdout
        correctly.
        """
        iterator = iter(self.reader)
        self.assertEqual(next(iterator), '{"a": {"B": 1},"b": 2}')
        self.assertEqual(next(iterator), '{"a": 1,"b": 2}')
        with self.assertRaises(StopIteration):
            next(iterator)

    def test_raises_err_if_non_0_return_code_with_stderr_from_bin(self):
        """Test that the `__next__` method raises a `BinaryExecutionException`
        if the binary returns a non-zero return code and there is stderr from
        the binary.
        """
        reader = bin_interface.BinaryReader("invalid_path")
        with self.assertRaises(BinaryExecutionException):
            next(iter(reader))
        reader.kill_subprocess_proc()

    def test_raises_err_if_non_0_return_code_no_stderr_from_bin(self):
        """Test that the `__next__` method raises a `BinaryExecutionException`
        if the binary returns a non-zero return code and there is no stderr
        from the binary.
        """
        reader = bin_interface.BinaryReader("invalid_path")
        with self.assertRaises(BinaryExecutionException):
            next(iter(reader))
        reader.kill_subprocess_proc()


class TestBinaryIterator(TestCase):
    """Tests for the `BinaryIterator` class."""

    def test__iter__returns_instance_of_self(self):
        """Test that the `__iter__` method returns an instance of itself."""
        iterator = bin_interface.BinaryIterator(None)
        self.assertIs(iter(iterator), iterator)

    def test__next__raises_stop_iter_if_no_stdout(self):
        """Test that the `__next__` method raises a `StopIteration` if there
        is no stdout.
        """
        iterator = bin_interface.BinaryIterator(
            SimpleNamespace(stdout=None, stderr=None, poll=lambda: 0)
        )
        with self.assertRaises(StopIteration):
            next(iterator)


class TestAsyncBinaryReader(ReaderInstanceMixin, IsolatedAsyncioTestCase):
    """Tests for the `AsyncBinaryReader` class."""

    @staticmethod
    def create_reader_instance():
        return bin_interface.AsyncBinaryReader(SAMPLE_DATA_PATH)

    async def test_ppopen_returns_popen(self):
        """Test that the `popen` method returns the correct type of
        instance.
        """
        proc = await self.reader.popen()
        self.assertIsInstance(proc, asyncio.subprocess.Process)

    async def test_raises_err_if_non_0_return_code(self):
        """Test that the `__anext__` method raises a `BinaryExecutionException`
        if the binary returns a non-zero return code.
        """

        async def fn():
            reader = bin_interface.AsyncBinaryReader("invalid_path")
            with self.assertRaises(BinaryExecutionException):
                async for _ in reader:
                    pass

        await asyncio.wait_for(fn(), timeout=0.1)

    async def test__aiter__returns_async_binary_iterator(self):
        """Test that the `__aiter__` method returns an `AsyncBinaryIterator`
        object.
        """
        aiter_ = self.reader.__aiter__()
        await aiter_.__anext__()

        self.assertIsInstance(aiter_, bin_interface.AsyncBinaryIterator)

    async def test__anext__valid(self):
        """Test that the `__anext__` method iterates over the binary stdout
        correctly.
        """
        iterator = self.reader.__aiter__()
        self.assertEqual(await iterator.__anext__(), '{"a": {"B": 1},"b": 2}')
        self.assertEqual(await iterator.__anext__(), '{"a": 1,"b": 2}')
        with self.assertRaises(StopAsyncIteration):
            await iterator.__anext__()


class TestAsyncBinaryIterator(ReaderInstanceMixin, IsolatedAsyncioTestCase):
    """Tests for the `AsyncBinaryIterator` class."""

    @staticmethod
    def create_reader_instance():
        return bin_interface.AsyncBinaryReader(SAMPLE_DATA_PATH)

    async def test_read_output_reads_stdout_til_exhaustion(self):
        """Test that the `read_output` method reads the stdout until it is
        exhausted.
        """
        proc = await self.reader.popen()
        self.assertEqual(
            await self.reader.read_output(proc),
            '{"a": {"B": 1},"b": 2}',
        )
        self.assertEqual(
            await self.reader.read_output(proc),
            '{"a": 1,"b": 2}',
        )
        self.assertEqual(await self.reader.read_output(proc), "")
        self.reader.kill_subprocess_proc()

    async def test_read_output_returns_empty_string_if_none_stdout(self):
        """Test that the `read_output` method returns an empty string if the
        stdout is `None`.
        """
        proc = await asyncio.create_subprocess_exec(
            "echo",
            "test",
            stdout=None,
        )
        self.assertEqual(await self.reader.read_output(proc), "")
        self.reader.kill_subprocess_proc()

    async def test__aiter__returns_instance_of_self(self):
        """Test that the `__aiter__` method returns an instance of itself."""
        iterator = bin_interface.AsyncBinaryIterator(None)
        self.assertIs(await iterator.__aiter__(), iterator)



import subprocess


class BinaryReader:
    def __init__(self, binary, fp):
        self.binary = binary
        self.file_path = fp
        self._proc = None


    def ppopen(self):
        self._proc = subprocess.Popen(
            [self.binary, self.file_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        # if self._proc.stderr:
        #     err = self._proc.stderr.readlines()

        while True:
            breakpoint()
            line = self._proc.stdout.readline().strip()
            # if self._proc.stderr:
            #     err = self._proc.stderr.readlines()

            if line == "" and self._proc.poll() is not None:
                break
            if line:
                print(line)

        







BinaryReader("./jsonl_converter", "/home/salaah/json-lineage/sample_data/sample.json").ppopen()


import os

SAMPLE_DATA_PATH = os.path.realpath(
    os.path.join(
        os.path.realpath(__file__),
        "..",
        "..",
        "..",
        "..",
        "sample_data",
        "sample.json",
    )
)



"""Contains the public API for the JSON lineage adapter. The module should
aim to mimic the `json` module in the Python standard library as much as
possible for easy adoption.
"""

from .bin_interface import AsyncBinaryReader, BinaryReader

__all__ = [
    "load",
    "aload",
]


def load(fp: str) -> BinaryReader:
    """Return a `BinaryReader` object for the given file path."""
    return BinaryReader(fp)


def aload(fp: str) -> AsyncBinaryReader:
    """Return an `AsyncBinaryReader` object for the given file path."""
    return AsyncBinaryReader(fp)
